#!/usr/bin/env python3

import numpy as np
import argparse
from argparse import RawTextHelpFormatter

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Print electronic energy, electronic energy with zero-point correction,\n"
                     "enthalpy, and Gibbs free energy from Gaussian output file(s) after frequency calculation\n"
                     "with some other structural information\n\n"
                     "Print the following values on terminal:\n"
                     "  File_name, Stoichiometry(*charge *multiplicity), E, E+ZPE, H, G, *number of imaginary frequencies\n"
                     "    *charge: printed with a sign (+ or -) if not charge-neutral\n"
                     "    *multiplicity: printed if not singlet\n"
                     "    *number of imaginary frequencies: printied if exist"
                     , formatter_class=RawTextHelpFormatter)

    parser.add_argument('file_name', nargs='+', help='Gaussian output file(s) (.out)')
    args = parser.parse_args()
    return args
    
def print_E(file_path):
    with open(file_path, 'r') as file:
        lines = file.readlines()

    stoich, E, EZPE, H, G, imagf = '', '', '', '', '', ''

    for line in lines:
        if "Stoichiometry" in line: 
            stoich = line.split()[1]
            break
    
    for idx, line in reversed(list(enumerate(lines))):
        if "SCF Done" in line: 
            E = line.split()[4]
            break
        
    for line in lines[len(lines)-idx:]:
        if "zero-point Energies" in line: EZPE = line.split()[6]
        if "thermal Enthalpies" in line: H = line.split()[6]
        if "thermal Free Energies" in line: G = line.split()[7]
        if "imaginary frequencies ignored" in line: imagf = line.split()[0]
        else: imagf = '0'
    
    return {
        'file_name': file_path.rsplit("/",1)[-1].rsplit(".",1)[0],
        'stoich': stoich,
        'E': E,
        'EZPE': EZPE,
        'H': H,
        'G': G,
        'imagf': imagf
    }

def main():
    args = parse_args()
    file_names = args.file_name

    for file_name in file_names:
        result = print_E(file_name)
        print(f"{result['file_name']}  {result['stoich']}  {result['E']}  {result['EZPE']}  "
              f"{result['H']}  {result['G']}  {result['imagf']}")

if __name__ == '__main__':
    main()
        


#!/usr/bin/env python3

import numpy as np
import pandas as pd
from periodictable import elements
import argparse
from argparse import RawTextHelpFormatter
from gaussianutility.utilities import readinput

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Sort geometry by an index; 'x', 'y', or 'z' coordinate, atomic number,\n"
                     "or ONIOM layer from low to high in the Gaussian input file (.com or .gjf)\n"
                     "Sorting by an atomic number is defualt"
                     "Sorting in ascending order is default\n\n"
                     "Return file_name.com: Gaussian input file with a sorted geometry",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument('file_name', help='Gaussian input file (.com or .gjf)')
    parser.add_argument('-s', '--sort', nargs='?', const=1, \
    help="Index for sorting;\n"+\
         "'x' or 'y' or 'z' for coordinate, 'A' for atomic number, and 'L' for ONIOM layer\n"+\
         "'L' can be combined with another index\n"+\
         "Ex) 'Lx' or 'xL' combines sorting by ONIOM layer and x coordinate in each layer\n"+\
         "'A' is default for non-ONIOM input and 'AL' is default for ONIOM input", default='AL')

    parser.add_argument('-o', '--order', nargs='?', const=1, \
    help="Index for the order for sorting;\n"+\
         "'a' for ascending order (default) and 'r' for reversed (descending) order\n"+\
         "Only one index can be provided\n"+\
         "If two sorting indeices (after -s) are provided for an ONIOM input,\n"+\
         "the ONIOM layer will be sorted from low to high automatically,\n"+\
         "and the atoms in each layer will be sorted based on the other index", default='a')

    args = parser.parse_args()
    return args
    
def sort_input(file_name, sortIdx, orderIdx):
    # Read a Gaussian input file
    route, title, charge_mult, df_geom, connectivity = readinput(file_name)

    oniomIdx = "oniom" in route.lower()
    
    # Check sorting and ordering indecies
    sortOptions = ['x', 'y', 'z', 'A', 'L', 'xL', 'Lx', 'yL', 'Ly', 'zL',\
    'Lz', 'AL', 'LA']

    if sortIdx not in sortOptions:
        raise ValueError("Unrecognized sorting index is provided")

    sortIdx = list(sortIdx)

    if not oniomIdx and 'L' in sortIdx:
        raise ValueError("The provided input structure does not have ONIOM partitions")
        
    if orderIdx not in ('a', 'r'):
        raise ValueError("Order index must be 'a for ascending or r for reversed (descending) order")
        
    # Set ordering index
    if len(sortIdx) == 1:
        if orderIdx == 'a': orderIdx = [1]
        elif orderIdx == 'r': orderIdx = [0]
        
    elif len(sortIdx) == 2:
        if orderIdx == 'a': orderIdx = [1, 1]
        elif orderIdx == 'r': orderIdx = [1, 0]

    # Sorting atoms
    ## Sorting based on the ONIOM layer first
    if oniomIdx and 'L' in sortIdx:
        df_layer_sort = pd.DataFrame({'ONIOM_layer': ['L', 'M', 'H']})
        layer_sort_mapping = df_layer_sort.reset_index().set_index('ONIOM_layer')
        df_geom['ONIOM_layer_num'] = df_geom['ONIOM_layer'].map(layer_sort_mapping['index'])
    
    ## If sorting only based on the ONION later
    if sortIdx == ['L']:
        df_geom = df_geom.sort_values(by=['ONIOM_layer_num'], ascending = orderIdx)
        df_geom = df_geom.drop(columns=['ONIOM_layer_num'])

    ## If sorting based on the atomic number
    elif 'A' in sortIdx:
        elem = np.array([[el.symbol, el.number] for el in elements])
        elem = {elem[i,0]: int(elem[i,1]) for i in range(len(elem))}
        df_geom['Atomic_num'] = [elem[atomSb] for atomSb in np.array(df_geom['Atom'])]

        if 'L' in sortIdx:
            df_geom = df_geom.sort_values(by=['ONIOM_layer_num', 'Atomic_num'], ascending = orderIdx)
            df_geom = df_geom.drop(columns=['ONIOM_layer_num', 'Atomic_num'])
        else:
            df_geom = df_geom.sort_values(by=['Atomic_num'], ascending = orderIdx)
            df_geom = df_geom.drop(columns=['Atomic_num'])
            
    ## If sorting based on the cartesian coordinate
    else:
        if 'x' in sortIdx: sortCoord = 'x'
        elif 'y' in sortIdx: sortCoord = 'y'
        elif 'z' in sortIdx: sortCoord = 'z'
            
        df_geom[sortCoord] = [float(val) for val in df_geom[sortCoord]]
             
        if 'L' in sortIdx:
            df_geom = df_geom.sort_values(by=['ONIOM_layer_num', sortCoord], ascending = orderIdx)
            df_geom = df_geom.drop(columns=['ONIOM_layer_num'])
        else:
            df_geom = df_geom.sort_values(by=sortCoord, ascending = orderIdx)
            
        df_geom[sortCoord] = ["{:.8f}".format(val) for val in df_geom[sortCoord]]

    # Modify connectivity data
    if 'connectivity' in route:
        atomIdxSwitch = np.vstack([df_geom.index.values, np.array(range(len(df_geom)))]).T
        atomIdxSwitch = atomIdxSwitch + np.ones(atomIdxSwitch.shape, dtype=int)
        atomIdxSwitch = {atomIdxSwitch[i,0]: atomIdxSwitch[i,1] for i in range(len(df_geom))}

        newConnect = []
        for line in connectivity:
            vals = line.split()
            for idx, val in enumerate(vals):
                try: vals[idx] = str(atomIdxSwitch[int(val)])
                except ValueError:
                    pass

            newConnect.append(vals)

        idx2remove = []
        connect2add = []
        for idx1, line in enumerate(newConnect):            
            if len(line) > 1:
                #1, 3, 5, ...
                for idx2 in range(len(line)):
                    if idx2 % 2 == 1:
                        connect2add.append([line[0], line[idx2], line[idx2+1]])
                                          
                idx2remove.append(idx1)
                                          
        for idx in sorted(idx2remove, reverse=True):
            del newConnect[idx]
                    
        newConnect = newConnect + connect2add

        for idx, line in enumerate(newConnect):
            if len(line) > 1 and int(line[0]) > int(line[1]):
                newConnect[idx] = [line[1], line[0], line[2]]
                
        missingAtom = np.setdiff1d(np.array(range(len(df_geom)))+1, np.unique([int(elem[0]) for elem in newConnect]))
        for elem in missingAtom:
            newConnect.append([str(elem)])
                
        newConnect = sorted(newConnect, key=lambda x:int(x[0]))
        newConnectSorted = []

        for i in range(len(df_geom)):
            sublist = [elem for elem in newConnect if int(elem[0]) == i+1]
            if len(sublist) > 1:
                starting = sublist[0][0]
                connected = []
                for elem in sublist:
                    connected.append(elem[1:])
                
                connected = [x for x in connected if x]
                
                if len(connected) > 1:
                    connected = [[starting]] + sorted(connected, key=lambda x:int(x[0]))
                else:
                    connected = [[starting]] + connected
            
                connected = [item for sublist in connected for item in sublist]
                newConnectSorted.append(connected)
                
            else:
                newConnectSorted.append(sublist[0])

    # Write Gaussian input file
    with open(file_name, 'w') as output:
        output.write(f"{route}\n")
        output.write(f"{title}\n")
        output.write(charge_mult)
        output.write(df_geom.to_csv(index=False, header=False, sep='\t'))
        output.write('\n\n')
        if 'newConnectSorted' in locals():
            for line in newConnectSorted:
                output.write(' '.join(line) + '\n')
            output.write('\n')

def main():
    args = parse_args()
    sort_input(args.file_name, args.sort, args.order)

if __name__ == "__main__":
    main()
    



#!/usr/bin/env python3

import numpy as np
import pandas as pd
import argparse
from argparse import RawTextHelpFormatter
from gaussianutility.utilities import readoutput

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Extract geometry from Gaussian output file (.out)\n"
                     "to Gaussian input file (.com)\n"
                     "This support ONIOM-type calculations\n"
                     "Connectivity information will be lost\n\n"
                     "Return file_name_geom.com: Gaussian input file",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument('file_name', help='Gaussian output file (.out)')
    parser.add_argument('-i', '--index', nargs='?', const=1, \
    help='Optimization step index;\n0 for input, -1 for the last geometry, '+\
         'and L for the lowest energy geometry', default=-1)
    args = parser.parse_args()
    return args
    
def main():
    args = parse_args()
    file_name, stepIdx = args.file_name, args.index

    # Read Gaussian output file
    routeStr, charge_mult, df_geom = readoutput(file_name, stepIdx)

    # Write .com file
    out_file = file_name.rsplit(".",1)[0] + "_geom.com"
    with open(out_file, 'w') as output:
        output.write(f"{routeStr}\n\n")
        output.write(f"{out_file}\n\n")
        output.write(f"{charge_mult}\n")
        output.write(df_geom.to_csv(index=False, header=False, sep='\t'))
        output.write("\n\n")

if __name__ == "__main__":
    main()
    


#!/usr/bin/env python3

import numpy as np
import pandas as pd
from periodictable import elements

"""

This file contains functions that can be used in the other scripts.
The utilities herein are not used by itself but inside another script.

"""

def readinput(file_name):
    """
    This reads an Gaussian inputfile and decomposes it into multiple elements 
    as a return for the use in other scripts.
    Return route section, title, charge and multiplicity, and and geometry.
    Geometry is a pandas dataframe and all the others are strings
    """
    # Check input file
    name, input_format = file_name.rsplit(".", 1)
    if input_format not in ('com', 'gjf'):
        raise TypeError('The input file format must be .com or .gjf')

    # Find the elements of the Gaussian input file
    with open(file_name, 'r') as inputfile:
        lines = inputfile.readlines()

    # Route section
    for idx, line in enumerate(lines):
        if "#" in line:
            idx_route = idx
            route = line
            break

    if len(lines[idx_route+1]) >=2:
        route = route.replace('\n', ' ')
        route += lines[idx_route+1]
        idx_route += 1

    connectIdx = "geom=connectivity" in route
            
    # Title and charge/multiplicity
    title = lines[idx_route+2]
    charge_mult = lines[idx_route+4]

    # Geometry
    geom=[]
    for idx, line in enumerate(lines[idx_route+5:]):
        geom.append(line.split())
        geomEndIdx = idx_route+idx+5
        if len(line) <= 2: break
        
    geom = list(filter(None, geom))
    
    connectivity = []
    if connectIdx:
        for idx, line in enumerate(lines[geomEndIdx:]):
            if line.startswith("1") or line.startswith(" 1"):
                break
        
        connectIdx = geomEndIdx + idx
        for line in lines[connectIdx:]:
            connectivity.append(line)
            if line.startswith(str(len(geom))) or line.startswith(" "+str(len(geom))):
                break

    if any(keyword in route.lower() for keyword in ["oniom"]):
        lineLen = 0
        for line in geom:
            if len(line) > lineLen: lineLen = len(line)

        if lineLen == 5:
            df_geom = pd.DataFrame(geom, columns = ['Atom','x','y','z','ONIOM_layer'])
        elif lineLen > 5:
            clmnName = []
            for i in range(lineLen):
                name = "C{}".format(i)
                clmnName.append(name)

            df_geom = pd.DataFrame(geom, columns = clmnName)
            
            if np.sum(np.mod(np.array(df_geom.iloc[:,1], dtype=float),1)) == 0:
                df_geom = df_geom.iloc[:,0:6]
                df_geom = df_geom.rename({'C0':'Atom', 'C1':'Index', 'C2':'x', 'C3':'y', 'C4':'z', 'C5':'ONIOM_layer'}, axis='columns')
            else:
                df_geom = df_geom.iloc[:,0:5]
                df_geom = df_geom.rename({'C0':'Atom', 'C1':'x', 'C2':'y', 'C3':'z', 'C4':'ONIOM_layer'}, axis='columns')
                
    else:
        lineLen = 0
        for line in geom:
            if len(line) > lineLen: lineLen = len(line)
        if lineLen == 4:
            df_geom = pd.DataFrame(geom, columns = ['Atom','x','y','z'])
        elif lineLen > 4:
            clmnName = []
            for i in range(lineLen):
                name = "C{}".format(i)
                clmnName.append(name)

            df_geom = pd.DataFrame(geom, columns = clmnName)
            
            if np.sum(np.mod(np.array(df_geom.iloc[:,1], dtype=float),1)) == 0:
                df_geom = df_geom.iloc[:,0:5]
                df_geom = df_geom.rename({'C0':'Atom', 'C1':'Index', 'C2':'x', 'C3':'y', 'C4':'z'}, axis='columns')
            else:
                df_geom = df_geom.iloc[:,0:4]
                df_geom = df_geom.rename({'C0':'Atom', 'C1':'x', 'C2':'y', 'C3':'z'}, axis='columns')            
                
    df_geom['Atom'] = np.array([atomSb.split('-')[0] for atomSb in list(df_geom['Atom'])])
    
    return route, title, charge_mult, df_geom, connectivity


def readoutput(file_name, stepIdx = -1):
    """
    This reads an Gaussian outputfile and decomposes it into multiple elements 
    as a return for the use in other scripts.
    Return route section, title, spin and multiplicity, and and geometry.
    Geometry is a pandas dataframe and all the others are strings
    """
    # Check input file
    name, input_format = file_name.rsplit(".", 1)
    if input_format != 'out':
        raise TypeError('The input file format must be .out')

    # Read file
    with open(file_name, 'r') as inputfile:
        lines = inputfile.readlines()

    # Read computational chemistry method and remove unnecessary keywords 
    for idx, line in enumerate(lines):
        if "#" in line:
            idx_route = idx
            break

    route = []
    for line in lines[idx_route:]:
        route.append(line.strip('\n'))
        if "----------------------------" in line:
            break

    route = route[:-1]
    for idx in range(len(route)):
        if route[idx].startswith(' '):
            route[idx] = route[idx][1:]

    routeStr = ""
    routeStr = routeStr.join(route)
    
    if 'qst' in routeStr:
        routeStr = routeStr.replace('=qst3', '').replace('qst3', '')
        routeStr = routeStr.replace('=qst2', '').replace('qst2', '')

    if "geom=connectivity" in routeStr:
        routeStr = routeStr.replace("geom=connectivity","")

    oniom = any(word in routeStr.lower() for word in ["oniom"])

    # Read charge and multiplicity information
    for idx, line in enumerate(lines):
        if "Charge =" in line:
            idx_charge = idx
            idx_charge_end = idx+1
            break

    if oniom:
        for idx, line in enumerate(lines[idx_charge:]):
            if not "Charge" in line:
                idx_charge_end = idx_charge+idx
                break

        charge_mult = []
        for line in lines[idx_charge: idx_charge_end]:
            charge_mult.append(str(line.split()[2]))
            charge_mult.append(str(line.split()[5]))

        charge_mult = " ".join(charge_mult)

    else:
        charge, multiplicity = lines[idx_charge].split()[2], lines[idx_charge].split()[5]
        charge_mult = "{} {}".format(charge, multiplicity)

    # Read oniom layer data
    iniGeom = []
    for line in lines[idx_charge_end:]:
        if len(line) <= 2: break
        iniGeom.append(line.split())

    iniGeom = list(filter(None, iniGeom))

    if oniom:
        oniomIdx = 0
        if len(iniGeom[0]) == 5:
            oniom_layer = [line[4] for line in iniGeom]
        elif len(iniGeom[0]) > 5:
            oniom_layer = [line[5] for line in iniGeom]

    # Read atom index data (0: optimized; -1:frozen)
    try:
        if isinstance(int(iniGeom[0][1]),int):
            indexFlag = True
    except ValueError:
        indexFlag = False
        
    if indexFlag:
        indices = [line[1] for line in iniGeom]

    # Read final geometry
    if stepIdx == "L":
        energies = []
        if oniom:
            for idx, line in enumerate(lines):
                if "ONIOM: extrapolated energy" in line:
                    energies.append(float(line.split()[-1]))
                    
            if energies[-1] == np.min(energies): stepIdx = -1
            else: stepIdx = np.argmin(np.array(energies))
            
        else:
            for idx, line in enumerate(lines):
                if "SCF Done:  " in line:
                    energies.append(float(line.split()[4]))
                    
            if energies[-1] == np.min(energies): stepIdx = -1
            else: stepIdx = np.argmin(np.array(energies))

    else: stepIdx = int(stepIdx)
    
    optLineNo = []
    standardOri = False
    for idx, line in enumerate(lines):
        if "Standard orientation" in line: 
            standardOri = True
            break
            
    for idx, line in enumerate(lines):        
        if standardOri:
            if "Standard orientation" in line:
                optLineNo.append(idx+1)
        else:
            if "Input orientation" in line:
                optLineNo.append(idx+1)

    if any(word in routeStr for word in ["freq","Freq","FREQ"]):
        optLineNo.pop(-1)

    geom = []
    for line in lines[optLineNo[stepIdx]+4:]:
        geom.append(line.split())
        if "----------------------------" in line:
            break

    # Convert geometry information to Pandas dataframe
    df_geom = pd.DataFrame(geom[:-1], columns = ['index','atomic number','atomic type','x','y','z'])
    df_geom = df_geom.drop(columns=['index', 'atomic type'])

    if oniom:
        df_geom['ONIOM_layer'] = oniom_layer
        
    if indexFlag:
        df_geom.insert(1, 'Index', indices)
        
    elem = np.array([[el.number, el.symbol] for el in elements])
    elem = {int(elem[i,0]): elem[i,1] for i in range(len(elem))}
    atomSb = [elem[atomNo] for atomNo in np.array(df_geom['atomic number'], dtype=int)]

    df_geom = df_geom.drop(columns='atomic number')
    df_geom.insert(0, 'Atom', atomSb)

    return routeStr, charge_mult, df_geom
    



#!/usr/bin/env python3

import numpy as np
import pandas as pd
import argparse
from argparse import RawTextHelpFormatter
from gaussianutility.utilities import readinput

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Freeze a given ONIOM layer(s) in a Gaussian input structure during\n"
                     "optimization calculation by adding indices (-1) next to atom symbols\n\n"
                     "Return file_name.com: Modified input structure",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument("file_name", help='Gaussian input file (.com or .gjf)')
    parser.add_argument('-i', '--index', nargs='?', const=1, \
    help="'H', 'M', or 'L' for freezing a high, medium, or low ONIOM layer, or their combination\n"+\
         "'U' unfreezes all the atoms. 'L' is default", default='L')
    args = parser.parse_args()
    return args
    
def freeze_layer(file_name, index):
    # Read a Gaussian input file
    route, title, charge_mult, df_geom, connectivity = readinput(file_name)

    if "oniom" not in route.lower():
        raise TypeError("Input structure must be in ONIOM scheme")
        
    if "Index" in df_geom.columns:
        df_geom["Index"] = 0
    else:
        df_geom.insert(1, "Index", 0)
        
    # Read index for freezing
    freezeIdx = list(freezeIdx)    
    if 'U' in freezeIdx:
        if len(freezeIdx) > 1:
            raise ValueError("U cannot be combined with other layer indices.")

    else:
        for layer in freezeIdx:
            if layer not in ['H', 'M', 'L']:
                raise ValueError("Layer index must be H, M, L, their combination, or U")
                        
            df_geom.loc[df_geom["ONIOM_layer"] == layer, "Index"] = -1

    # Write Gaussian input file
    with open(file_name, 'w') as output:
        output.write(f"{route}\n")
        output.write(f"{title}\n")
        output.write(charge_mult)
        output.write(df_geom.to_csv(index=False, header=False, sep='\t'))
        output.write('\n\n')
        output.writelines(connectivity)
        output.write('\n')

def main():
    args = parse_args()
    freeze_layer(args.file_name, args.index)

if __name__ == "__main__":
    main()
    




#!/usr/bin/env python3

import numpy as np
import math as m
import sys
import argparse
from argparse import RawTextHelpFormatter

# Thermodynamic constants
kB = 1.380649e-23 # J/K
h = 6.62607015e-34 # Js
Rgas = 8.3144626 # J/mol/K
Na = 6.02214076e23

def parse_args():
    parser = argparse.ArgumentParser(
        description= """
        Calculate Gibbs free energy (in Hartrees) at different tempeature(s) from Gaussian output file (.out)
        The Gaussian job must be normally terminated with frequency calculation
        Print temperatures and Gibbs free energies on terminal.
        """,
        formatter_class=RawTextHelpFormatter)

    parser.add_argument('file_name', help='Gaussian output file (.out)')
    parser.add_argument('T1', type=float, help='Temperature to calculate G (in K)')
    parser.add_argument('T2', type=float, nargs='?', help='Upper bound of temperature range (in K)')
    parser.add_argument('step_number', type=int, nargs='?', help="Number of steps of temperature between T1 and T2 to calculate G", default=20)
    args = parser.parse_args()
    return args
    
# Calculate contributions from translational, rotational, vibrational, and eletronic motions
# The equations are sourced from "Thermochemistry in Gaussian" written by Joseph W. Ochterski (2000)
# https://gaussian.com/wp-content/uploads/dl/thermo.pdf
def Contributions(temp, mass, press, vibTemp, multiplicity, rho_r, theta_r):
    ## Translational part
    qt = m.pow(2*m.pi*mass*kB*temp/h**2, 3/2) * kB*temp/press
    St = Rgas*(m.log(qt)+1+3/2)
    Et = 3/2*Rgas*temp

    ## Rotational part
    if len(theta_r) > 1:    
        qr = m.pi**(1/2)/rho_r*m.sqrt((temp**3/np.prod(theta_r)))
    elif len(theta_r) == 1:
        qr = m.pi**(1/2)/rho_r*m.sqrt((temp/(theta_r[0])))
        
    if 'qr' in locals():
        Sr = Rgas*(m.log(qr)+3/2)
        Er = 3/2*Rgas*temp
    else:
        Sr = 0
        Er = 0

    ## Vibrational part
    summ1 = 0 # for entropy contribution
    summ2 = 0 # for energy contribution
    for val in vibTemp:
        summ1 += val/temp/(m.exp(val/temp) - 1) - m.log(1-m.exp(-val/temp))
        summ2 += val*(1/2 + 1/(m.exp(val/temp)-1))

    Sv, Ev = Rgas*np.array([summ1, summ2])

    ## Electronic part
    Se = Rgas * m.log(multiplicity)
    Ee = 0
    Stot = St + Sr + Sv + Se
    Etot = Et + Er + Ev + Ee
    
    return Stot, Etot # return total entropy and energy corrections
    
def gibbs_temp(file_name, T1, T2, step_number):
    #Set temperature based on the provided arguments
    if T2:
        temperature = np.linspace(T1, T2, step_number)
    else:
        temperature = np.array([T1])

    # Exctract thermochemistry result from the output file
    with open(file_name, 'r') as inFile:
        lines = inFile.readlines()

    multiplicity = None
    thermochem = []
    collecting_thermochem = False

    for line in lines:
        if "Multiplicity" in line:
            multiplicity = int(line.split()[5]) 
        elif "Thermochemistry" in line:
            collecting_thermochem = True
        elif "Total Bot" in line:
            break
        elif collecting_thermochem:
            thermochem.append(line.strip())
    
    if not thermochem:
        raise ValueError("Required information not found in the output file.")

    # Extract required informations from the thermochemistry results
    for idx, line in enumerate(thermochem):
        if "Pressure" in line:
            press = float(line.split()[4])*101325 # Pa
        elif "Molecular mass" in line:
            mass = float(line.split()[2])*1.6605391e-27 # kg
        elif "Rotational symmetry number" in line:\
            rho_r = float(line.split()[3]) 
        elif "Rotational temperatures" in line: 
            if len(line.split()) < 5: 
                theta_r = np.array([float(line.split()[3])])
            else:
                theta_r = np.float64(np.array(line.split()[3:6]))
        elif "Vibrational temperature" in line:
            vibTempBeginIdx = idx
        elif "Zero-point correction" in line: 
            zpCorr = float(line.split()[2])*2625.4996394799e3 # J/mol
            vibTempEndIdx = idx
        elif "Sum of electronic and zero-point Energies" in line:
            ElectE = float(line.split()[6])*2625.4996394799e3 - zpCorr # J/mol

    vibTemp = []
    for line in thermochem[vibTempBeginIdx:vibTempEndIdx]:
        for elem in line.split():
            try: vibTemp.append(float(elem))
            except: ValueError
    
    vibTemp = np.array(vibTemp)

    # Arrays of entropy and energy corrections in the given temperature (range)
    StotArr=[]
    EtotArr=[]
    if 'rho_r' not in locals():
        rho_r = []
        theta_r = []

    for temp in temperature:
        Stot, Etot = Contributions(temp, mass, press, vibTemp, multiplicity, rho_r, theta_r)
        StotArr.append(Stot)
        EtotArr.append(Etot)

    StotArr = np.array(StotArr)
    EtotArr = np.array(EtotArr)
    
    Hcorr = EtotArr + kB*temperature*Na # Thermal corrections to Enthalpy
    Gcorr = Hcorr - temperature * StotArr # Thermal corrections to Free energy
    Gibbs = (ElectE + Gcorr) / 2625.4996394799e3 # Final Gibbs free energy in Hartree

    temperature = np.ndarray.tolist(np.round(temperature,decimals=6))
    Gibbs = np.ndarray.tolist(np.round(Gibbs,decimals=6))

    print("Temperature [K]: " + str(temperature).strip('[]'))
    print("Gibbs free energy [Hartree]: " + str(Gibbs).strip('[]'))

def main():
    args = parse_args()
    gibbs_temp(args.file_name, args.T1, args.T2, args.step_number)

if __name__ == "__main__":
    main()
    



#!/usr/bin/env python3

import pandas as pd
import numpy as np
from periodictable import elements
import argparse
from argparse import RawTextHelpFormatter
from gaussianutility.utilities import readinput

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Convert Gaussian input file to .vasp file\n\n"
                     "Return file_name.vasp: VASP geometry file (POSCAR)",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument("file_name", help='Gaussian input file (.com or .gjf)')
    return parser.parse_args()

def com_2_vasp(file_name):
    # Read geometry from a Gaussian input file
    _, _, _, df_geom, _ = readinput(file_name)
    df_geom = df_geom[['Atom', 'x', 'y', 'z']]
    df_geom['Atom'] = np.array([i.rsplit("-",1)[0] for i in list(df_geom['Atom'])])
    if not 'Tv' in df_geom['Atom'].values:
        raise ValueError('The Gaussian input file does not contain lattice information')

    # Convert Gaussian input to VASP input
    elem = [el.symbol for el in elements]
    df_atom_sort = pd.DataFrame(elem[1:], columns = ['symbol'])
    atom_sort_mapping = df_atom_sort.reset_index().set_index('symbol')
    df_geom['Atomic_num'] = df_geom['Atom'].map(atom_sort_mapping['index'])

    df_geom = df_geom.sort_values(by=['Atomic_num'], ascending = 0)
    df_geom = df_geom.drop(columns=['Atomic_num'])

    # Read lattice information
    lattice = df_geom[-3:].drop(['Atom'], axis=1) 

    df_geom = df_geom.drop(df_geom[df_geom['Atom'] == 'Tv'].index)
    df_geom = df_geom.drop(df_geom[df_geom['Atom'] == '?s'].index)

    atomlist = np.array(df_geom['Atom'])
    coord = df_geom[['x','y','z']]

    uniqatomlist = []
    uniqatomcount = []

    for elem in atomlist:
        if elem not in uniqatomlist:
            uniqatomlist.append(elem)
            uniqatomcount.append(np.count_nonzero(atomlist == elem))

    uniqatomlist = np.array(uniqatomlist)
    uniqatomcount= np.array(uniqatomcount, dtype=str)

    # Write VASP input file
    name = ".".join(file_name.rsplit(".",1)[0:-1])
    out_file = name + ".vasp"
    with open(out_file, 'w') as output:
        output.write(name + '\n')
        output.write('1.0\n')
        output.write(lattice.to_csv(index=False, header=False, sep='\t'))
        output.write('   '+' '.join(uniqatomlist)+'\n')
        output.write('   '+' '.join(uniqatomcount)+'\n')
        output.write('Cartesian\n')
        output.write(coord.to_csv(index=False, header=False, sep='\t'))
        output.write('\n')

def main():
    args = parse_args()
    com_2_vasp(args.file_name)

if __name__ == "__main__":
    main()





#!/usr/bin/env python3

import numpy as np
import pandas as pd
import argparse
from argparse import RawTextHelpFormatter
from gaussianutility.utilities import readinput

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Convert Gaussian input file (.com or .gjf) to xyz file\n\n"
                     "Return file_name.xyz: xyz structure file",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument("file_name", help='Gaussian input file (.com or .gjf)')
    return parser.parse_args()
    
def com_2_xyz(file_name):
    # Read geometry from a Gaussian input file
    _, _, _, df_geom, _ = readinput(file_name)
    df_geom = df_geom[['Atom', 'x', 'y', 'z']]
    df_geom['Atom'] = np.array([i.rsplit("-",1)[0] for i in list(df_geom['Atom'])])
    elem_len = str(len(df_geom))

    # Write xyz file
    name = ".".join(file_name.rsplit(".",1)[0:-1])
    out_file = name + ".xyz"
    with open(out_file, 'w') as output:
        output.write(elem_len + '\n')
        output.write(out_file + '\n')
        output.write(df_geom.to_string(index=False, header=False) + '\n')

def main():
    args = parse_args() 
    com_2_xyz(args.file_name)

if __name__ == "__main__":
    main()




#!/usr/bin/env python

import numpy as np
import itertools
import mendeleev as md
import pandas as pd
import argparse
from argparse import RawTextHelpFormatter

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Convert vasp input file to Gaussian input file (.com)\n"
                     "The computational method in the Gaussian calculation must be modified manually if needed\n\n"
                     "Return file_name.com: Gaussian input file",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument("file_name", help='VASP input file (.vasp)')
    return parser.parse_args()

def vasp_2_com(file_name):
    # Read geometry information and make Pandas dataframe
    lines = open(file_name).readlines()

    elem_type = lines[5].split()
    elem_num = np.array(list(map(int, lines[6].split())))
    elect_num = np.array([md.element(e).electrons for e in elem_type])

    multiplicity = 1 if sum(elect_num * elem_num) % 2 == 0 else 2

    elem_list = np.array(list(itertools.chain.from_iterable([ [e] * n for e, n in zip(elem_type, elem_num)])))
    
    geom = np.array([line.split() for line in lines[8:] if len(line) >= 2], dtype='float_')

    lattice_scailing_factor = float(lines[1].split()[0])
    lattice_vector = [line.split() for line in lines[2:5]]
    lattice_vector = np.array(lattice_vector, dtype = float)

    geomMod = lines[7].split()[0]
    if geomMod.lower() == 'cartesian':
        geom *= lattice_scailing_factor
    elif geomMod.lower() == 'direct':
        geom = np.dot(lattice_vector, geom.T).T * lattice_scailing_factor
        
    df_geom = pd.DataFrame(geom, columns = ['x','y','z'])
    df_geom.insert(loc=0, column='element', value=elem_list)

    # Write .com file
    out_file = file_name.rsplit(".",1)[0] + ".com"
    with open(out_file, 'w') as output:
        output.write('# pbepbe/3-21g/auto\n\n')
        output.write(f'{file_name.rsplit(".", 1)[0]}\n\n')
        output.write(f'0 {multiplicity}\n')
        output.write(df_geom.to_string(index=False, header=False) + '\n')
        output.write('Tv ' + lines[2])
        output.write('Tv ' + lines[3])
        output.write('Tv ' + lines[4])
        output.write('\n')

def main():
    args = parse_args()
    file_name = args.file_name

    if not file_name.endswith(".vasp"):
        raise ValueError('The input structure must be a .vasp file')

    vasp_2_com(file_name)

if __name__ == "__main__":
    main()




#!/usr/bin/env python3

import numpy as np
import argparse
from argparse import RawTextHelpFormatter
import mendeleev as md

def parse_args():
    parser = argparse.ArgumentParser(
        description= "Convert xyz file to Gaussian input file (.com)\n\n"
                     "Return file_name.com: Gaussian input file",
        formatter_class=RawTextHelpFormatter)

    parser.add_argument("file_name", help='XYZ structure file (.xyz)')
    return parser.parse_args()

def xyz_2_com(file_name):
    # Read geometry from XYZ file
    lines = open(file_name).readlines()

    # Find the line where the XYZ coordinates start
    for idx, line in enumerate(lines[:]):
        if len(line.split()) == 4:
            try:
                _ = [float(val.replace('n','')) for val in line.split()[1:]]
                break
            except ValueError:
                continue

    # Decide multiplicity assuming overall neutral charge
    elemt_list = [line.split()[0] for line in lines[idx:]]
    elemt_type, elemt_count = np.unique(elemt_list, return_counts=True)
    elemt_electrons = [md.element(elemt).electrons for elemt in elemt_type]
    total_electrons = sum(np.array(elemt_electrons) * elemt_count)
    multiplicity = 1 if total_electrons % 2 == 0 else 2

    # Write Gaussian input file
    name = ".".join(file_name.rsplit(".",1)[0:-1])
    out_file = name + ".com"
    with open(out_file, 'w') as output:
        output.write('# hf/3-21g\n\n')
        output.write(f'{out_file}\n\n')
        output.write(f'0 {multiplicity}\n')
        for line in lines[idx:]:
            output.write(line)
        output.write('\n')

def main():
    args = parse_args()
    file_name = args.file_name

    if not file_name.endswith(".xyz"):
        raise ValueError('The input structure must be a .xyz file')
    
    xyz_2_com(file_name)

if __name__ == "__main__":
    main()





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gaussianutility",
    version="0.1.6",
    author="Sungil Hong",
    author_email="s.hong@pitt.edu",
    description="Useful utilities to use Gaussian09/16 software",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        'console_scripts': [
            'com2vasp=gaussianutility.com2vasp:main',
            'com2xyz=gaussianutility.com2xyz:main',
            'xyz2com=gaussianutility.xyz2com:main',
            'freezeLayer=gaussianutility.freezeLayer:main',
            'gibbsTemp=gaussianutility.gibbsTemp:main',
            'out2com=gaussianutility.out2com:main',
            'printE=gaussianutility.printE:main',
            'sortInput=gaussianutility.sortInput:main',
            'spectrum=gaussianutility.spectrum:main',
            'vasp2com=gaussianutility.vasp2com:main'
        ],
    },
    url="https://github.com/Sungil-Hong/gaussianutility",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"],
    python_requires='>=3.6',
    install_requires=['numpy>=1.17.2',
                      'pandas>=1.1.0',
                      'periodictable>=1.6.1',
                      'argparse>=1.1',
                      'matplotlib>=3.5.1',
                      'mendeleev>=0.9.0']
)



#!/usr/bin/env python3

import math as m
import numpy as np
import pandas as pd
import argparse
from argparse import RawTextHelpFormatter
import matplotlib.pyplot as plt

##### Change these lines to modify X range or make corrections #####
##### For UV-Vis #####
UV_wl_range = np.linspace(100,500,801) # nm
UV_alpha = 1 #UV-Vis wavelength scaling factor; default
#UV_alpha = 1.12 #for M062X; ref: https://doi.org/10.26434/chemrxiv-2022-pfvhc-v3
UV_delta = 0.4/m.sqrt(2) #UV-Vis bandwith parameter, eV; default
#UV_delta = 0.21  #for M062X; ref: https://doi.org/10.26434/chemrxiv-2022-pfvhc-v3

##### For IR and Raman #####
IR_wv_range = np.linspace(0,4000,8001) # cm-1
IR_wn_factor = 1 # default
#IR_wn_factor = 0.947 #For M062X; ref: J. Phys. Chem. A, 121(11), 2265 (2017)#
IR_HWHM = 4 # IR peak half-width at half-max, cm-1 # default

# Define functions to extract spectrum data from output file(s)
def uv_vis(file_name):
    readfile = open(file_name).readlines()
    wavelengths = []
    strengths = []
    
    for idx, line in enumerate(reversed(readfile)):
        if "Excitation energies and oscillator strengths" in line:
            break
            
    ex_idx = len(readfile)-idx-1
    
    for line in readfile[ex_idx:]:
        if 'Excited State' in line:
            wavelengths.append(float(line.split()[6]))
            strengths.append(float(line.split()[8].split('=')[1]))
            
    if len(wavelengths) == 0:
        raise TypeError('The Gaussian job does not look like excited state calculations')

    wavelengths = np.array(wavelengths)
    strengths = np.array(strengths)
    
    return wavelengths, strengths


def ir(file_name):
    readfile = open(file_name, 'r')
    frequencies = []
    intensities = []
    
    for line in readfile:
        if 'Frequencies --' in line:
            frequencies.append(line.split()[2:])
        if 'IR Inten' in line:
            intensities.append(line.split()[3:])
            
    if len(intensities) == 0:
        raise TypeError('The Gaussian job does not look like conntaining normal IR information')
            
    frequencies = np.array(frequencies, dtype='float').flatten() #cm-1
    intensities = np.array(intensities, dtype='float').flatten() #km/mole

    return frequencies, intensities

def raman(file_name):
    readfile = open(file_name, 'r')
    frequencies = []
    intensities = []
    
    for line in readfile:
        if 'Frequencies --' in line:
            frequencies.append(line.split()[2:])
        if 'Raman Activ' in line:
            intensities.append(line.split()[3:])
            
    if len(intensities) == 0:
        raise TypeError('The Gaussian job does not look like conntaining normal IR information')
            
    frequencies = np.array(frequencies, dtype='float').flatten() #cm-1
    intensities = np.array(intensities, dtype='float').flatten() 

    return frequencies, intensities

# Constants
h = 6.6261e-34 # J/s
c = 299792458  # m/s
J2eV = 6.02214076e23/96.48530749925793/1000

# Converstion between energy and wavelength
def wvl2E(wvl): # wvl in nm
    E = h*c/(wvl/1e9)*J2eV
    return E # E in eV

def E2wvl(E): #E in eV
    wvl = h*c*1e9/(E/J2eV)
    return wvl # wvl in nm

# Gaussian distribution for UV-Vis
def uvGauss(x, f, wv):
    #Reference: https://gaussian.com/uvvisplot/
#    return 13.062973*f*3099.6*m.exp(-((1/x-1/wv)/0.00032262)**2)
    return 13.062973*f*E2wvl(m.sqrt(2)*UV_delta)*m.exp(-((wvl2E(x)-wvl2E(wv)/UV_alpha)/UV_delta/m.sqrt(2))**2)

# Cauchy distribution for normal and raman IR
def cauchy(x, mu):
    gamma = IR_HWHM
    return 1/m.pi/gamma/(1+((x-mu)/gamma)**2)

def parse_args():
    parser = argparse.ArgumentParser(
        description = """
        Produce Normal-IR, Raman, or UV-Vis spectrum from Gaussian output file (.out)
        For Raman spectrum, the Gaussian calculation should be done with freq=raman keyword
        For UV-Vis spectrum, excited state calculations (like TD-DFT and EOM-CCSD) should be done
        Linear combination of multiple spectra for mixtures is available by adding ratio arguments
        This case, the sum of the ratios must be 1

        Examples of command line usage are:
            spectrum ir file1.out
            spectrum uv file1.out file2.out -r 0.8 0.2

        Return file_name_type.png: Spectrum image
               file_name_type.csv: Sepctrum data
        """,
        formatter_class=RawTextHelpFormatter
    )

    parser.add_argument('type', 
                        choices=["uv", "ir", "raman"],
                        help="Type of spectrum; 'uv' for UV-Vis, 'ir' for normal IR, and 'raman' for Raman")
    parser.add_argument('file_name', nargs='+', 
                        help="Gaussian output files (.out)")
    parser.add_argument('-r', '--ratio', nargs='?', 
                        help="Ratios of input structures (required when multiple structures are provided)")
    args = parser.parse_args()

    return args

def main():
    # Read provided arguments
    args = parse_args()
    type_name = args.type
    file_names = args.file_name
    ratios = [float(val) for val in args.ratio] if args.ratio else [1]

    if len(file_names) != len(ratios):
        raise ValueError("Ratio(s) of one or more species is not provided")
    
    if sum(ratios) > 1.0001 or sum(ratios) < 0.9999:
        raise ValueError('Sum of the ratios of the species should be 1')

    names = []
    for file_name in file_names:
        name, informat = file_name.rsplit(".", 1)
        names.append(name)
        if informat not in ("out"):
            raise TypeError("The input file format must be .out")

    # Plot spectra
    # UV-Vis
    if type_name == 'uv':
        spectrum_save = '_'.join(names) + '_uv_vis.png'
        data_save = '_'.join(names) + '_uv_vis.csv'
        
        f = plt.figure()
        f.set_figwidth(12)
        f.set_figheight(8)
        
        xrange = UV_wl_range
        curve_per_file = []
        
        for file_name in file_names:
            wavelengths, strengths = uv_vis(file_name)
            
            curve_per_x = []
            for idx, wv in enumerate(wavelengths):
                wv_array = np.ones(len(xrange)) * wv
                st_array = np.ones(len(xrange)) * strengths[idx]
                
                curve = np.array(list(map(uvGauss, xrange, st_array, wv_array)))
                curve_per_x.append(curve)
            
            final_curve = np.sum(curve_per_x, axis=0)
            curve_per_file.append(final_curve)
            
            if len(file_names) == 1:
                plt.plot(xrange, final_curve, 'b')
            else:
                blue = 0.8
                alpha = 0.6
                red = np.random.uniform(low=0, high=1)
                green = np.random.uniform(low=0, high=1)
                plt.plot(xrange, final_curve, color = (red, green, blue), alpha=alpha)
        
        if len(file_names) == 1:
            f.legend(names, loc='upper right', bbox_to_anchor=(0.9, 0.88))
            
        else:
            summed_curve = np.dot(np.array(curve_per_file).T, np.array([ratios]).T)
            plt.plot(xrange, summed_curve, 'b')
            f.legend(names + ['Mixture'], loc='upper right', bbox_to_anchor=(0.9, 0.88))
            
        plt.xlabel('Wavelength, $\lambda$  $(nm)$', fontsize=20)
        plt.ylabel('Absorbance $(L/mol/cm)$', fontsize=20)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.savefig(spectrum_save)
        
        if len(file_names) == 1:
            uv_vis_df = pd.DataFrame(
                np.hstack((np.array([xrange]).T,np.array(curve_per_file).T)),
                columns = ['wavelength'] + names)
        else:
            uv_vis_df = pd.DataFrame(
                np.hstack((np.array([xrange]).T,np.array(curve_per_file).T, summed_curve)),
                columns = ['wavelength']+ names + ['Mixture'])
        uv_vis_df.to_csv(data_save, index=False)

    # IR: both normal IR and raman
    if type_name in ['ir', 'raman']:
        if type_name == 'ir':
            spectrum_save = '_'.join(names) + '_ir.png'
            data_save = '_'.join(names) + '_ir.csv'
        elif type_name == 'raman':
            spectrum_save = '_'.join(names) + '_raman.png'
            data_save = '_'.join(names) + '_raman.csv'
        
        f = plt.figure()
        f.set_figwidth(12)
        f.set_figheight(8)
        
        xrange = IR_wv_range
        curve_per_file = []
        
        for file_name in file_names:
            if type_name == 'ir':
                frequencies, intensities = ir(file_name)
                frequencies *= IR_wn_factor # Correction factor
                intensities *= 1e5/2.24e3
            elif type_name == 'raman':
                frequencies, intensities = raman(file_name)
                
            curve_per_x = []
            for idx, mu in enumerate(frequencies):
                mu_array = np.ones(len(xrange)) * mu
                curve = np.array(list(map(cauchy, xrange, mu_array))) * intensities[idx]
                curve_per_x.append(curve)
            
            final_curve = np.sum(curve_per_x, axis=0)
            curve_per_file.append(final_curve)
            
            if len(file_names) == 1:
                plt.plot(xrange, final_curve, 'b')
            else:
                blue = 0.8
                alpha = 0.6
                red = np.random.uniform(low=0, high=1)
                green = np.random.uniform(low=0, high=1)
                plt.plot(xrange, final_curve, color = (red, green, blue), alpha=alpha)
        
        if len(file_names) == 1:
            f.legend(name, loc='upper right', bbox_to_anchor=(0.9, 0.88))
            
        else:
            summed_curve = np.dot(np.array(curve_per_file).T, np.array([ratios]).T)
            plt.plot(xrange, summed_curve, 'b')
            f.legend(names + ['Mixture'], loc='upper right', bbox_to_anchor=(0.9, 0.88))
            
        plt.xlabel('Wavenumber, $\\nu$  $(cm^{-1})$', fontsize=20)
        plt.ylabel('Absorbance (a.u.)', fontsize=20)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.savefig(spectrum_save)
        
        if len(file_names) == 1:
            ir_df = pd.DataFrame(
                np.hstack((np.array([xrange]).T,np.array(curve_per_file).T)),
                columns = ['wavenumber'] + names)
        else:
            ir_df = pd.DataFrame(
                np.hstack((np.array([xrange]).T,np.array(curve_per_file).T, summed_curve)),
                columns = ['wavenumber'] + names + ['Mixture'])
        ir_df.to_csv(data_save, index=False)

if __name__ == "__main__":
    main()
    
